/*
 * Sprt Micronics
 *
 * CUPS Filter
 *
 * [ Linux ]
 * compile cmd: gcc -Wl,-rpath,/usr/lib -Wall -fPIC -O2 -o rastertoprinter rastertoprinter.c -lcupsimage -lcups
 * compile requires cups-devel-1.1.19-13.i386.rpm (version not neccessarily important?)
 * find cups-devel location here: http://rpmfind.net/linux/rpm2html/search.php?query=cups-devel&submit=Search+...&system=&arch=
 *
 * [ Mac OS X ]
 * compile cmd: gcc -Wall -fPIC -o rastertoprinter rastertoprinter.c -lcupsimage -lcups -arch ppc -arch i386 -DMACOSX
 * compile requires cupsddk-1.2.3.dmg (version not neccessarily important?)
 * find cupsddk location here: http://www.cups.org/software.php
 */

/*
 * Copyright (C) 2004-2009 Sprt Micronics Co., Ltd.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * Include necessary headers...
 */

#include <cups/cups.h>
#include <cups/ppd.h>
#include <cups/raster.h>
#include <stdlib.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>

#define MAX(a,b) ( ((a) > (b)) ? (a) : (b) )

#define FALSE 0
#define TRUE  (!FALSE)

struct settings_
{
    int modelNumber;
    float pageWidth;
    float pageHeight;
    int pageCutType;
    int docCutType;
    int cashDrawer1;
    int cashDrawer2;
    int bytesPerScanLine;
};

struct command
{
    int    length;
    char* command;
};

// define printer initialize command
static const struct command printerInitializeCommand =
{2,(char[2]){0x1b,0x40}};

// define raster mode start and stop command
//send 0x15,0x01, before exit rasterMode. so rasterModeCommand[0] length is 5.
static const struct command rasterModeCommand[2] =
{{5,(char[5]){0x15,0x01,0x1d,0x44,0x00}},
{3,(char[3]){0x1d,0x44,0x01}}};

// define cash drawer commands
static const struct command cashDrawerCommand [2] =
{{5,(char[5]){0x1b,0x70,0x00,0x32,0x32}},                    // 0 open drawer 1
 {5,(char[5]){0x1b,0x70,0x01,0x32,0x32}}};                   // 1 open drawer 2

// define page cut type commands
static const struct command cutTypeCommand [3] =
{{5,(char[5]){0x1d,0x56,0x63,0x03,0xa8}},                    // 0 no cut, tear bar       
 {5,(char[5]){0x1d,0x56,0x63,0x04,0x20}},                    // 1 half cut
 {5,(char[5]){0x1d,0x56,0x63,0x05,0x20}}};                   // 2 full cut

// define end page command
static const struct command cutCommand =
{3,(char[3]){0x1d,0x56,0x6d}};

inline void outputCommand(struct command output)
{
    int i = 0;

    for (; i < output.length; i++)
    {
        putchar(output.command[i]);
    }
}

inline int getOptionChoiceIndex(const char * choiceName, ppd_file_t * ppd)
{
    ppd_choice_t * choice;
    ppd_option_t * option;

    choice = ppdFindMarkedChoice(ppd, choiceName);
    if (choice == NULL)
    {
        if ((option = ppdFindOption(ppd, choiceName))          == NULL) return -1;
        if ((choice = ppdFindChoice(option,option->defchoice)) == NULL) return -1;
    }

    return atoi(choice->choice);
}

inline void getPageWidthPageHeight(ppd_file_t * ppd, struct settings_ * settings)
{
    ppd_choice_t * choice;
    ppd_option_t * option;

    char width[20];
    int widthIdx;

    char height[20];
    int heightIdx;

    char * pageSize;
    int idx;

    int state;

    choice = ppdFindMarkedChoice(ppd, "PageSize");
    if (choice == NULL)
    {
        option = ppdFindOption(ppd, "PageSize");
        choice = ppdFindChoice(option,option->defchoice);
    }

    widthIdx = 0;
    memset(width, 0x00, sizeof(width));

    heightIdx = 0;
    memset(height, 0x00, sizeof(height));

    pageSize = choice->choice;
    idx = 0;

    state = 0; // 0 = init, 1 = width, 2 = height, 3 = complete, 4 = fail

    while (pageSize[idx] != 0x00)
    {
        if (state == 0)
        {
            if (pageSize[idx] == 'X')
            {
                state = 1;

                idx++;
                continue;
            }
        }
        else if (state == 1)
        {
            if ((pageSize[idx] >= '0') && (pageSize[idx] <= '9'))
            {
                width[widthIdx++] = pageSize[idx];

                idx++;
                continue;
            }
            else if (pageSize[idx] == 'D')
            {
                width[widthIdx++] = '.';

                idx++;
                continue;
            }
            else if (pageSize[idx] == 'M')
            {
                idx++;
                continue;
            }
            else if (pageSize[idx] == 'Y')
            {
                state = 2;

                idx++;
                continue;
            }
        }
        else if (state == 2)
        {
            if ((pageSize[idx] >= '0') && (pageSize[idx] <= '9'))
            {
                height[heightIdx++] = pageSize[idx];

                idx++;
                continue;
            }
            else if (pageSize[idx] == 'D')
            {
                height[heightIdx++] = '.';

                idx++;
                continue;
            }
            else if (pageSize[idx] == 'M')
            {
                state = 3;
                break;
            }
        }

        state = 4;
        break;
    }

    if (state == 3)
    {
        settings->pageWidth = atof(width);
        settings->pageHeight = atof(height);
    }
    else
    {
        settings->pageWidth = 0;
        settings->pageHeight = 0;
    }
}

inline void initializeSettings(char * commandLineOptionSettings, struct settings_ * settings)
{
    ppd_file_t *    ppd         = NULL;
    cups_option_t * options     = NULL;
    int             numOptions  = 0;

    ppd = ppdOpenFile(getenv("PPD"));

    ppdMarkDefaults(ppd);

    numOptions = cupsParseOptions(commandLineOptionSettings, 0, &options);
    if ((numOptions != 0) && (options != NULL))
    {
        cupsMarkOptions(ppd, numOptions, options);
        cupsFreeOptions(numOptions, options);
    }

    memset(settings, 0x00, sizeof(struct settings_));

    settings->modelNumber = ppd->model_number;

    settings->pageCutType                    = getOptionChoiceIndex("PageCutType"                   , ppd);
    settings->docCutType                     = getOptionChoiceIndex("DocCutType"                    , ppd);
    settings->cashDrawer1                    = getOptionChoiceIndex("CashDrawer1"                   , ppd);
    settings->cashDrawer2                    = getOptionChoiceIndex("CashDrawer2"                   , ppd);

    getPageWidthPageHeight(ppd, settings);
    settings->bytesPerScanLine = settings->pageWidth;

    ppdClose(ppd);
}

void startJob(struct settings_ settings)
{
    //init printer.1B 40
    outputCommand(printerInitializeCommand);

    //enter raster model.
    outputCommand(rasterModeCommand[1]);
  
    //set cut enable. 0-no cut, 1-half cut.
    if ((settings.pageCutType > 0) || (settings.docCutType > 0))
    {
        outputCommand(cutTypeCommand[1]);
    }

    //CashDrawer1 and CashDrawer2 open or not
    if(settings.cashDrawer1 == 1) // 0-off, 1-before, 2-after
    {
        outputCommand(cashDrawerCommand[0]);
    }

    if(settings.cashDrawer2 == 1) // 0-off, 1-before, 2-after
    {
        outputCommand(cashDrawerCommand[1]);
    }
}

void endPage(struct settings_ settings)
{
    //page cut
    if(settings.pageCutType > 0)
    {
        outputCommand(cutCommand);
    }
}

void endJob(struct settings_ settings)
{
    if(settings.pageCutType <= 0 && settings.docCutType > 0)
    {
        outputCommand(cutCommand);
    }

    //CashDrawer1 and CashDrawer2 open or not
    if(settings.cashDrawer1 == 2) // 0-off, 1-before, 2-after
    {
        outputCommand(cashDrawerCommand[0]);
    }

    if(settings.cashDrawer2 == 2) // 0-off, 1-before, 2-after
    {
        outputCommand(cashDrawerCommand[1]);
    }

    //exit raster model.
    outputCommand(rasterModeCommand[0]);
}

#define CLEANUP                                                         \
{                                                                       \
    if (originalRasterDataPtr   != NULL) free(originalRasterDataPtr);   \
    cupsRasterClose(ras);                                               \
    if (fd != 0)                                                        \
    {                                                                   \
        close(fd);                                                      \
    }                                                                   \
}

int main(int argc, char *argv[])
{
    int                 fd                      = 0;        /* File descriptor providing CUPS raster data */
    cups_raster_t *     ras                     = NULL;     /* Raster stream for printing*/
    cups_page_header2_t header;                              /* CUPS Page header*/

    unsigned char *     rasterData              = NULL;     /* Pointer to raster data buffer */
    unsigned char *     originalRasterDataPtr   = NULL;     /* Copy of original pointer for freeing buffer*/
    int                 scanLineBlank           = 0;        /* Set to TRUE is the entire scan line is blank (no black pixels)*/
    int                 lastBlackPixel          = 0;        /* Position of the last byte containing black pixels in the scan line*/
    int                 numBlankScanLines       = 0;        /* Number of scanlines that were entirely black */
    struct settings_    settings;                           /* Configuration settings*/

    int page = 0;  /* Current page*/
    int y = 0;  /* Vertical position in page 0 <= y <= header.cupsHeight*/
    int i = 0;


    if (argc < 6 || argc > 7)
    {
	    fputs("ERROR: rastertoprinter job-id user title copies options [file]\n", stderr);
	    return EXIT_FAILURE;
    }

    if (argc == 7)
    {
	    if ((fd = open(argv[6], O_RDONLY)) == -1)
	    {
	        perror("ERROR: Unable to open raster file - ");
	        sleep(1);
	        return EXIT_FAILURE;
	    }
    }
    else
    {
	    fd = 0;
    }

    //usb print will lost one package data, so send one package at first.
    int k = 0;
    for (k = 0; k < 64; k ++)
    {
        putchar((char)0x00);
    }

    initializeSettings(argv[5], &settings);
    startJob(settings);
    ras = cupsRasterOpen(fd, CUPS_RASTER_READ);
    page = 0;

    while (cupsRasterReadHeader2(ras, &header))
    {
        if ((header.cupsHeight == 0) || (header.cupsBytesPerLine == 0))
        {
            break;
        }

        if (rasterData == NULL)
        {
            rasterData = malloc(header.cupsBytesPerLine);
            if (rasterData == NULL)
            {
                CLEANUP;
                return EXIT_FAILURE;
            }
         //memset(rasterData, 0x00, header.cupsBytesPerLine);
        }

        originalRasterDataPtr = rasterData;  // used to later free the memory
        //pageSetup(settings, header);

        page++;
        fprintf(stderr, "PAGE: %d %d\n", page, header.NumCopies);
        numBlankScanLines = 0;

        if (header.cupsBytesPerLine <= settings.bytesPerScanLine)
        {
            settings.bytesPerScanLine = header.cupsBytesPerLine;
        }

        for (y = 0; y < header.cupsHeight; y ++)
        {
            if ((y & 127) == 0)
            {
                fprintf(stderr, "INFO: Printing page %d, %d%% complete...\n", page, (100 * y / header.cupsHeight));            
            }

            if (cupsRasterReadPixels(ras, rasterData, header.cupsBytesPerLine) == 0)
            {
               break;
            }

            for (i = settings.bytesPerScanLine - 1; i >= 0; i--)
            {
                if (((char) *(rasterData + i)) != ((char) 0x00))
                {
                    break;
                }
            }

            if (i == -1)
            {
                scanLineBlank = TRUE;
                numBlankScanLines++;
            }
            else
            {
                lastBlackPixel = i + 1;
                scanLineBlank = FALSE;
            }

            if (scanLineBlank == FALSE)
            {
                if (numBlankScanLines > 0)
                {
                    putchar((char)0x15);
                    putchar((char)numBlankScanLines);
                    numBlankScanLines = 0;
                }

                putchar((char)0x16);
                putchar((char)lastBlackPixel);
                for (i = 0; i < lastBlackPixel; i++)
                {
                    putchar((char) *rasterData);
                    rasterData++;
                }
            }
            rasterData = originalRasterDataPtr;
        }
        endPage(settings);
    }

    endJob(settings);
    CLEANUP;

    if (page == 0)
    {
        fputs("ERROR: Garland No pages found!\n", stderr);
        return EXIT_FAILURE;
    }
    else
    {
        fputs("INFO: Ready to print.\n", stderr);
        return EXIT_SUCCESS;
    }
}

// end of rastertoprinter.c
